import re
from itertools import izip
from ...data.materials.util.LookUpData import LookUpData

class CompositionEntry(object):
    """Class that defines a CompositionEntry object.

    Mainly used to store ids, names and fractions of elements belonging to a
    single compound.

    Attributes
    ----------
    lp_element_names : array-like
        Names of each element. A list of string values.
    lp_sorting_order : array-like
        Rank of each element (used in display order). A list of int values.
    element_ids : array-like
        Element ids present in composition. A list of int values.
    element_names : array-like
        Element names present in composition. A list of string values.
    fractions : array-like
        Fraction of each element. A list of float values.
    number_in_cell : float
        Number of atoms in cell (used to convert when printing).

    """

    # Names of each element.
    lp_element_names = LookUpData.element_names

    # Rank of each element (used in display order).
    lp_sorting_order = LookUpData.sorting_order

    # Element ids present in composition.
    element_ids = []

    # Element names present in composition.
    element_names = []

    # Fraction of each element.
    fractions = []

    # Number of atoms in cell (used to convert when printing).
    number_in_cell = -float("inf")

    def __init__(self, composition=None, element_ids=None,
                 element_names=None, fractions=None):
        """Function to make a new instance by paring the composition of an
        object, provided by a string or list of element ids/names and fractions.

        Parameters
        ----------
        composition : str
            The chemical formula of a material.
        element_ids : array-like
            List of integers denoting the element ids.
        element_names : array-like
            List of strings denoting the element names.
        fractions : array-like
            List of floats denoting the element fractions.

        Raises
        ------
        RuntimeError
            If composition parsing was unsuccessful.

        Notes
        -----
        First splits using the regex [A-Z][^A-Z] to separate each component
        of the alloy. This assumes that capitalization is used properly. Then
        splits each component into the alphabetical part (assumed to be
        element name) and the numeric part (amount of element). If no numeric
        part is present, 1 is assumed. An element can appear more than once.

        """

        # Parse composition if passed.
        if composition:
            comp_map = self.parse_composition(composition)

            # Crash if nothing is read.
            if not comp_map:
                raise RuntimeError("No composition was read.")

            # Store values.
            self.element_ids = comp_map.keys()
            self.element_names = [self.lp_element_names[i] for i in
                                  self.element_ids]
            self.fractions = comp_map.values()
            self.sort_and_normalize()
        elif fractions and (element_names or element_ids):
            self.set_composition(fractions, element_ids=element_ids,
                                 element_names=element_names)

    def parse_composition(self, composition):
        """Function to parse a string containing the composition.

        Supports parentheses and addition compounds (ex: Na_2CO_3-10H_2O).
        Note, will not properly parse addition compounds inside parentheses
        (ex: Na_2(CO_3 - 10H_2O)_1).

        Parameters
        ----------
        composition : str
            The chemical formula of a material.

        Returns
        -------
        output : dict
            Dictionary containing element ids and fractions as keys and
            values respectively.

        Raises
        ------
        ValueError
            If closing parenthesis is missing.
            If parenthesis is not recognized.

        """

        # Check for a guest structure (ex: Al2O3-2H20).
        start_guest = -1 if "-" not in composition else composition.index("-")
        pos = -1 if chr(183) not in composition else composition.index(chr(183))

        if pos != -1 and (pos < start_guest or start_guest == -1):
            start_guest = pos

        # If this composition doesn't contain a guest either.
        if start_guest == -1:

            # Check for ({['s
            start_paren = -1 if '(' not in composition else composition.index(
                '(')
            pos = -1 if '{' not in composition else composition.index(
                '{')
            if pos != -1 and (pos < start_paren or start_paren == -1):
                start_paren = pos
            pos = -1 if '[' not in composition else composition.index(
                '[')
            if pos != -1 and (pos < start_paren or start_paren == -1):
                start_paren = pos

            # If none found, parse as normal.
            if start_paren == -1:
                return self.parse_element_amounts(composition)

            # First, get the part of string before the parens.
            first_part = composition[:start_paren]

            # Next, find the portion inside of the parens.
            paren_type = composition[start_paren]
            pos = start_paren
            paren_balance = 1
            while paren_balance > 0:
                # Increment position.
                pos += 1

                # Check whether we are past the end of the string.
                if pos > len(composition):
                    raise ValueError("Missing close paren. Start: "
                                     ""+composition[start_paren:start_paren+3])

                # Get the character at the current position.
                cur_char = composition[pos]

                if paren_type == '(':
                    if cur_char == '(':
                        paren_balance += 1
                    elif cur_char == ')':
                        paren_balance -= 1
                elif paren_type == '{':
                    if cur_char == '{':
                        paren_balance += 1
                    elif cur_char == '}':
                        paren_balance -= 1
                elif paren_type == '[':
                    if cur_char == '[':
                        paren_balance += 1
                    elif cur_char == ']':
                        paren_balance -= 1
                else:
                    raise ValueError("Unrecognized paren type: "+paren_type)

            # Get the multiplier of the composition inside the parens.
            end_paren = pos
            pos += 1
            mult = ""
            while pos < len(composition) and (composition[pos].isdigit() or
                                                  composition[pos] == "."):
                mult += composition[pos]
                pos += 1

            paren_mult = 1.0 if not mult else float(mult)

            # Get the portion insides of those parens, and end portion.
            inside_paren = composition[start_paren+1:end_paren]
            end_portion = composition[pos:]

            total_comp = self.parse_element_amounts(first_part)
            add_comp = self.parse_composition(inside_paren)

            self.combine_compositions(total_comp, add_comp, paren_mult)

            add_comp = self.parse_composition(end_portion)

            self.combine_compositions(total_comp, add_comp, 1.0)
            return total_comp

        else:
            # Find how many guests.
            end_host = start_guest
            pos = start_guest + 1
            mult = ""
            while pos < len(composition) and (composition[pos].isdigit() or
                                                  composition[pos] == "."):
                mult += composition[pos]
                pos += 1
            start_guest = pos
            guest_mult = 1.0 if not mult else float(mult)

            # Split compound.
            host_part = composition[:end_host]
            guest_part = composition[start_guest:]

            # Compute them separately.
            host_comp = self.parse_composition(host_part)
            guest_comp = self.parse_composition(guest_part)
            self.combine_compositions(host_comp, guest_comp, guest_mult)
            return host_comp

    def set_composition(self, amounts, element_ids=None, element_names=None,
                        to_sort=True):
        """Function to set the composition of this entry.

        Checks to make sure all elements have positive amounts.

        Parameters
        ----------
        amounts : array-like
            List of amounts (float) for each element.
        element_ids : array-like
            List of element ids (integers).
        element_names : array-like
            List of element names (strings).

        Raises
        ------
        ValueError
            If either element names or ids are missing.
            If lists have different lengths.

        """

        if element_ids is None and element_names is None:
            raise ValueError("Must include a list of element names or ids "
                             "along with amounts!")
        if element_names is None:
            if len(amounts) != len(element_ids):
                raise ValueError("Array lengths of amounts and element_ids "
                                 "must be equal!")
            self.element_ids = element_ids
        else:
            if len(amounts) != len(element_names):
                raise ValueError("Array lengths of amounts and element_names "
                                 "must be equal!")
            self.element_names = element_names
            self.element_ids = [self.lp_element_names.index(elem) for elem in
                                element_names]
        self.fractions = amounts
        self.sort_and_normalize(to_sort)

    def __copy__(self):
        """Function to make a copy of this instance.

        Returns
        -------
        x : CompositionEntry
            A copy of this instance.

        """

        x = type(self)(self.__class__)
        x.__dict__.update(self.__dict__)
        x.element_ids = list(self.element_ids)
        x.fractions = list(self.fractions)
        return x

    def parse_element_amounts(self, composition):
        """Function to compute fractions of element given a string of elements
        and amounts.

        Parameters
        ----------
        composition : str
            Composition of a material.

        Returns
        -------
        tmp_entry : dict
            Dictionary containing element ids and fractions as keys and
            values respectively.

        Raises
        ------
        ValueError
            If either element names or ids are not recognized.
            If element amount is not a number.

        """

        # Create temporary entry.
        tmp_entry = {}

        # Add up all the constituents.
        comp_iter = re.compile(r"[A-Z][^A-Z]*").finditer(composition)
        elem_pattern = re.compile(r"[A-Z][a-z]?")
        fraction_pattern = re.compile(r"[.0-9]+")
        for comp in comp_iter:
            component = comp.group()

            # Get the element information.
            elem_matcher = elem_pattern.match(component)
            if not elem_matcher:
                raise ValueError("Something has gone horribly wrong!")
            element = elem_matcher.group()
            if element == "D" or element == "T":
                element = "H"
            if element not in self.lp_element_names:
                raise ValueError("Element "+element+" not recognized")
            element_id = self.lp_element_names.index(element)

            # Get the amount of this element.
            fraction_matcher = fraction_pattern.search(component)
            fraction = 1.0
            if fraction_matcher:
                try:
                    fraction = float(fraction_matcher.group())
                except ValueError:
                    raise ValueError("Element amount "+
                        fraction_matcher.group()+" not a valid number.")

            # Skip if fraction is zero.
            if fraction == 0:
                continue

            if element_id in tmp_entry:
                tmp_entry[element_id] += fraction
            else:
                tmp_entry[element_id] = fraction

        return tmp_entry


    def get_element_names(self):
        """Function to get the element names in the composition.

        Returns
        -------
        element_names : array-like
            List of element names (strings).
        """

        return self.element_names

    def get_element_ids(self):
        """Function to get the element ids in the composition.

        Returns
        -------
        element_ids : array-like
            List of element ids (int).
        """

        return self.element_ids

    def get_element_fractions(self):
        """Function to get the element fractions in the composition.

        Returns
        -------
        element_fractions : array-like
            List of element fractions (float).

        """

        return self.fractions

    def get_element_fraction(self, name=None, id=None):
        """Function to get the element fraction given either the name or the id
        of the element.

        Parameters
        ----------
        name : str
            Name of the element.
        id   : int
            Id of the element.

        Returns
        -------
        fraction : float
            Elemental fraction.

        """

        e_id = id
        if e_id is None:
            for i in range(len(self.lp_element_names)):
                if self.lp_element_names[i].lower() == name.lower():
                    e_id = i

        for i in range(len(self.element_ids)):
            if self.element_ids[i] == e_id:
                return self.fractions[i]
        return 0.0

    def __cmp__(self, other):
        """Function to compare two CompositionEntry objects.

        Parameters
        ----------
        other : CompositionEntry
            Other composition entry to compare.

        Returns
        -------
        output : int
            -1 if self < other , 1 if self > other or 0 if self = other.

        """

        if isinstance(other, CompositionEntry):
            # If this has more elements, this is greater.
            if len(self.element_ids) != len(other.element_ids):
                return -1 if len(self.element_ids) < len(other.element_ids) \
                    else 1
            # Check which has greater element fractions.
            for i in range(len(other.element_ids)):
                if self.element_ids[i] != other.element_ids[i]:
                    return -1 if self.element_ids[i] < other.element_ids[i] \
                        else 1
                elif self.fractions[i] != other.fractions[i]:
                    return -1 if self.fractions[i] < other.fractions[i] else 1
            # We have concluded that they are equal.
        return 0

    def __hash__(self):
        """Function to compute the hashcode of this instance.

        Computes the hashcode of the list of element ids and fractions
        separately. Then does the logical XOR operation between the two
        hashcodes and 1 and returns the result.

        Returns
        -------
        output : int
            Hashcode of this instance.

        """

        h1 = h2 = 0
        for e,f in izip(self.element_ids, self.fractions):
            h1 += 31*h1 + hash(e)
            h2 += 31*h2 + hash(f)

        return h1 ^ h2 ^ 1

    def __eq__(self, other):
        """Function to compare the equality between two CompositionEntry
        instances.

        Parameters
        ----------
        other : CompositionEntry
            Other composition entry to compare.

        Returns
        -------
        output : bool
            True if they are equal and False otherwise.

        """

        if isinstance(other, CompositionEntry):
            if len(self.element_ids) != len(other.element_ids):
                return False
            return self.element_ids == other.element_ids and \
                   self.element_names == other.element_names and \
                   self.fractions == other.fractions
        return False

    def sort_and_normalize(self, to_sort=True):
        """Function to sort the element ids based on their electronegativity
        order and normalizes the fractions.

        Makes sure the entry is in a proper format. Must be run from
        constructor.

        Parameters
        ----------
        to_sort : bool
            Whether to sort as well as normalize or just normalize this
            instance.

        """

        # Sort elements based on the electronegativity order.
        if to_sort:
            tmp_tuple = zip(self.element_ids, self.fractions)
            tmp_tuple.sort(key=lambda (x,y): self.lp_sorting_order[x])
        else:
            tmp_tuple = zip(self.element_ids, self.fractions)

        # Normalize the fractions.
        self.number_in_cell = sum(self.fractions)

        self.element_ids = []
        self.element_names = []
        self.fractions = []

        for e_id, f in tmp_tuple:
            if f > 0.0:
                f_ = float(f) / self.number_in_cell
                if e_id not in self.element_ids:
                    self.element_ids.append(e_id)
                    self.element_names.append(self.lp_element_names[e_id])
                    self.fractions.append(f_)
                else:
                    idx = self.element_ids.index(e_id)
                    self.fractions[idx] += f_

    def combine_compositions(self, total_comp, add_comp, multiplier):
        """Function to add one CompositionEntry to another.

        Parameters
        ----------
        total_comp : dict
            Dictionary containing element ids and fractions as keys and
            values respectively. Composition to be added to.
        add_comp   : dict
            Dictionary containing element ids and fractions as keys and
            values respectively. Composition to add.
        multiplier : float
            Factor to multiply with.

        Returns
        -------

        """

        for e,f in add_comp.iteritems():
            # If total_comp contains this element.
            if e in total_comp:
                total_comp[e] += multiplier * f
            else:
                total_comp[e] = multiplier * f

    @classmethod
    def print_number(self, fraction, n_in_formula_unit):
        """Function to print out the number of atoms in a formula unit for each
        element given its fraction.

        Parameters
        ----------
        fraction : array-like
            List of element fractions (floats) to be printed.
        n_in_formula_unit : int
            Number of atoms in a formula unit.

        Returns
        -------
        output : str
            Formatted fractions.

        """

        output = []
        for f in fraction:
            expanded = f * n_in_formula_unit
            if abs(expanded - round(expanded)) < 0.0001:
                output.append("" if int(expanded) == 1 else "{0:d}".format(
                    int(expanded)))
            else:
                output.append("{0:1.3f}".format(expanded))
        return output

    def __str__(self):
        """Function to print the composition entry in the proper format.

        Returns
        -------
        output : str
            Correctly formatted output.

        """

        if not self.element_names:
            return "Elem_list not defined"
        else:
            output = ""
            numbers = self.print_number(self.fractions, self.number_in_cell)
            for i in range(len(self.element_names)):
                output += self.element_names[i]
                output += "" if not numbers else numbers[i]

            return output

    @classmethod
    def import_composition_list(self, file_path):
        """Function to read a list of compositions from a file.

        Parameters
        ----------
        file_path : str
            Path to the file containing the list of compositions.

        Returns
        -------
        composition_list : array-like
            A list of CompositionEntry's corresponding to the file contents.

        """
        composition_list = []
        with open(file_path, 'r') as f:
            for line in f.readlines():
                words = line.strip()
                entry = CompositionEntry(composition=words)
                composition_list.append(entry)

        return composition_list

    @classmethod
    def import_values_list(self, file_path):
        """Function to read a list of target property values from a file.

        Target property values are used to develop machine learning models.

        Parameters
        ----------
        file_path : str
            Path to the file containing the list of compositions.

        Returns
        -------
        property_list : array-like
            A list of target property values (floats) corresponding to the
            file contents.

        """

        property_list = []
        with open(file_path, 'r') as f:
            for line in f.readlines():
                property_list.append(float(line.strip()))

        return property_list


import types
import numpy as np
from ....data.materials.CompositionEntry import CompositionEntry

class CompositionDistanceFilter:
    """Class to filter compositions based on distance from a target composition.
    Filters any composition where the maximum change in any element is less
    than a certain value.

    Attributes
    ----------
    target_composition : CompositionEntry
        Composition of the target.
    threshold : float
        Threshold to be used.


    """
    def __init__(self):
        """Initialize field variables here.
        """

        # Target composition.
        self.target_composition = None

        # Threshold distance.
        self.threshold = 0.0

    def set_target_composition(self, entry):
        """Function to define the target composition.

        Parameters
        ----------
        entry : CompositionEntry
            Desired target composition.

        """
        self.target_composition = entry

    def set_distance_threshold(self, distance):
        """Function to define the threshold composition distance. Here distance
        is defined as the maximum change in the fraction of any element.

        Parameters
        ----------
        distance : float
            Target threshold in %


        """
        self.threshold = distance / 100.0

    @classmethod
    def compute_distance(self, entry_1, entry_2, p):
        """Function to compute the distance between two entries. Distance is
        defined as the L_p norm of the distance between element fractions.

        Parameters
        ----------
        entry_1 : CompositionEntry
            Composition of entry 1.
        entry_2 : CompositionEntry
            Composition of entry 2.
        p : int
            Desired norm.

        Returns
        -------
        output : float
            Distance between two entries.

        """

        # Get the set of common elements with fractions greater than zero to
        # consider.
        elements = set()
        for e1 in entry_1.get_element_ids():
            elements.add(e1)

        for e2 in entry_2.get_element_ids():
            elements.add(e2)

        # Compute differences
        dist = 0.0
        for e in elements:
            diff = entry_1.get_element_fraction(id=e) - \
                   entry_2.get_element_fraction(id=e)
            if p == 0:
                if diff != 0:
                    dist += 1
            elif p == -1:
                dist = max(dist, diff)
            else:
                dist += abs(diff)**p

        # Compute distance
        if p == 0 or p == -1:
            return dist

        return dist**(1.0 / p)

    def label(self, entries):
        """Function to compute labels of composition entries indicating whether
        or not they are within the threshold of the target composition.

        Parameters
        ----------
        entries : array-like
            A list of CompositionEntry's.

        Returns
        -------
        label : array-like
            A numpy array containing True if the entry is within bounds of
            the target composition and False otherwise.

        Raises
        ------
        ValueError
            If input is not of type list.
            If items in the list are not CompositionEntry instances.
        """

        # Raise exception if input argument is not of type list of
        # CompositionEntry's.
        if (type(entries) is not types.ListType):
            raise ValueError("Argument should be of type list of "
                             "CompositionEntry's.")
        elif (entries and not isinstance(entries[0], CompositionEntry)):
            raise ValueError("Argument should be of type list of "
                             "CompositionEntry's.")

        target_elements = self.target_composition.get_element_ids()
        target_fractions = self.target_composition.get_element_fractions()

        label = np.array([False]*len(entries))
        for i,entry in enumerate(entries):
            within_bounds = True
            for e in range(len(target_elements)):
                v1 = entry.get_element_fraction(id=target_elements[e])
                if abs(v1 - target_fractions[e]) > self.threshold:
                    within_bounds = False
                    break
            label[i] = within_bounds

        return label


import types
from heapq import heappush, heappop
import numpy as np
import pandas as pd
from ....data.materials.CompositionEntry import CompositionEntry
from ....data.materials.util.LookUpData import LookUpData
from ....data.utilities.filters.CompositionDistanceFilter import \
    CompositionDistanceFilter
from ....utility.EqualSumCombinations import EqualSumCombinations

class APEAttributeGenerator:
    """Class to compute features using Atomic Packing Efficiency (APE) of
    nearby clusters.

    Attributes
    ----------
    packing_threshold : float
        Threshold at which to define a cluster as efficiently packed.
    n_nearest_to_eval : array_like
        Number of nearest clusters to assess. An array or list of int values.
        Default : [1, 3, 5]
    max_n_types : int
        Maximum number of types over which to search for clusters.
    radius_property : str
        Name of elemental property to use as atomic radius.

    Notes
    -----
    The APE, as defined by Laws et al. [1], is determined based
    on the ideal and actual ratio between the central and shell atoms of an
    atomic cluster with a certain number of atoms. Often, the packing
    efficiency is described as a ratio between these two quantities:

    [Packing efficiency] = [Ideal radius ratio] / [Actual radius ratio]

    The ideal ratio is determined based on the ratio between the size of a
    central atom and the neighboring atoms such that the packing around the
    central atom is maximized. These optimal ratios for clusters for
    different numbers of atoms have been tabulated by Miracle et al. [2].

    The actual ratio is computed by dividing the radius of the central atom
    by the average of the central atoms.

    We currently use this framework to create two types of features:

    Distance to nearest clusters with a packing efficiency better than a
    certain threshold. If there are fewer than a requested number of
    efficiently packed clusters in an alloy system, the average is taken to
    be the average distance to all of the clusters. These features are
    designed to measure the availability of efficiently-packed atomic
    configurations in the liquid.

    Mean packing efficiency of the system assuming that the composition of
    the first nearest neighbor shell is equal to the composition of the
    system. Each atom type is surrounded by the number of atoms that
    maximizes the packing efficiency. As shown in recent work by Laws et al.
    [1], bulk metallic glasses are known to form when the clusters around all
    types of atom have the same composition as the alloy and are efficiently
    packed. We compute the average APE for each atom in the system, under this
    assumption, and the average deviation from perfect packing.

    Advanced Notes
    --------------
    This algorithm currently evaluates all possible clusters provided a list
    of elements. As the number of clusters scales with N!, the runtime of
    this algorithm scales with N!.

    For now, we only search for clusters with up to 7 atoms (max_n_types) in
    order to avoid this combinatorial problem. In practice, the algorithm
    picks the top 7 alloys with the highest fractions. While not idea,
    this might work in practice since most alloys have fewer than 7 main
    components. Many alloys >10 elements in the specification, but many are
    impurities that may not be present in large enough amounts to really
    affect the determination of efficiently packed clusters.

    References
    ----------
    .. [1] K. J. Laws, D. B. Miracle, and M. Ferry, "A predictive structural
    model for bulk metallic glasses," Nature Communications, vol. 6, p. 8123,
    Sep. 2015.
    .. [2] D. B. Miracle, E. A. Lord, and S. Ranganathan, "Candidate Atomic
    Cluster Configurations in Metallic Glass Structures," MATERIALS
    TRANSACTIONS, vol. 47, no. 7, pp. 1737 -- 1742, 2006.
    .. [3] D. B. Miracle, D. V. Louzguine-Luzgin, L. V. Louzguina-Luzgina,
    and A. Inoue, "An assessment of binary metallic glasses: correlations
    between structure, glass forming ability and stability," International
    Materials Reviews, vol. 55, no. 4, pp. 218--256, Jul. 2010.

    """

    # Threshold at which to define a cluster as efficiently packed.
    # Packing efficiency is defined by |APE - 1|. Default value for this
    # parameter is 0.01.
    packing_threshold = 0.01

    # Number of nearest clusters to assess.
    n_nearest_to_eval = [1, 3, 5]

    # Maximum number of types over which to search for clusters. If an
    # alloy has more than this number of elements, the code will only
    # search for clusters with the most prevalent elements.
    max_n_types = 6

    # Name of elemental property to use as atomic radius. By default,
    # uses the radii from Ref [3] (see references section above).
    radius_property = "MiracleRadius"

    def set_packing_threshold(self, threshold):
        """Function to define the threshold at which a cluster is considered
        efficiently packed.

        Parameters
        ----------
        threshold : float
            Desired threshold. Default: 0.01

        Raises
        ------
        ValueError
            If threshold value is negative.

        """

        if threshold < 0:
            raise ValueError("Threshold must be positive.")
        self.packing_threshold = threshold

    def set_n_nearest_to_eval(self, values):
        """Function to define the number of nearest neighbor clusters to
        evaluate when computing features.

        Parameters
        ----------
        values : array-like
            Number of nearest clusters to assess. An array or list of int
            values.

        """

        self.n_nearest_to_eval = values

    def set_radius_property(self, prop):
        """Function to set the name of the elemental property used to define
        radii.

        By default uses the "MiracleRadius" property which is from an
        assessment by Miracle et al. [2].

        Parameters
        ----------
        prop : str
            Name of property used to define radii.

        """

        self.radius_property = prop

    @classmethod
    def compute_APE(self, n_neighbors=None, center_radius=None,
                    neigh_eff_radius=None, radii=None, center_type=None,
                    shell_types=None):
        """Function to compute the APE of a cluster, given the identities of
        the central and 1st neighbor atoms or just the number of neighbors and
        the radii.

        Here, we follow the formulation given by Laws et al. [1].

        APE = ideal radius ratio / (radius of central atom / effective radius
        of nearest neighbors)

        Parameters
        ----------
        n_neighbors : int
            Number of 1st nearest neighbors in the cluster.
        center_radius : float
            Radius of the central atom.
        neigh_eff_radius : float
            Effective radius of the 1st shell. Usually computed as the
            average radius of all atoms in the shell.
        radii : array-like
            Radius of each atom type. A list of float values.
        center_type : int
            Type of atom in the center.
        shell_types : array-like
            Number of atoms of each type in the outer shell. Must be same
            length as radii.

        Returns
        -------
        output : float
            APE, as defined in the function description.

        """
        n_n = n_neighbors
        c_r = center_radius
        n_e_r = neigh_eff_radius

        if n_n is None and c_r is None and n_e_r is None:
            # Get the radius of the central atom.
            c_r = radii[center_type]

            # Get the mean radius of the 1st neighbor shell.
            # n_e_r = np.dot(shell_types, radii)
            # n_n = np.sum(shell_types)
            n_e_r = 0.0
            n_n = 0.0
            for i in xrange(len(shell_types)):
                n_e_r += shell_types[i] * radii[i]
                n_n += shell_types[i]


            # for i in xrange(len(shell_types)):
            #     n_neighbors += shell_types[i]
            #     neighbor_radius += shell_types[i] * radii[i]

            n_e_r /= n_n

        #  The ideal radius ratio is only known for clusters with 3 and 24 (
        # inclusive) neighbors. If you request outside of this range,
        # the value of 3 is set for anything less than 3 and the value of 24
        # is set for anything larger than 24.
        if n_n <= 3:
            ideal_ratio = 0.154701
        elif n_n == 4:
            ideal_ratio = 0.224745
        elif n_n == 5:
            ideal_ratio = 0.361654
        elif n_n == 6:
            ideal_ratio = 0.414213
        elif n_n == 7:
            ideal_ratio = 0.518145
        elif n_n == 8:
            ideal_ratio = 0.616517
        elif n_n == 9:
            ideal_ratio = 0.709914
        elif n_n == 10:
            ideal_ratio = 0.798907
        elif n_n == 11:
            ideal_ratio = 0.884003
        elif n_n == 12:
            ideal_ratio = 0.902113
        elif n_n == 13:
            ideal_ratio = 0.976006
        elif n_n == 14:
            ideal_ratio = 1.04733
        elif n_n == 15:
            ideal_ratio = 1.11632
        elif n_n == 16:
            ideal_ratio = 1.18318
        elif n_n == 17:
            ideal_ratio = 1.24810
        elif n_n == 18:
            ideal_ratio = 1.31123
        elif n_n == 19:
            ideal_ratio = 1.37271
        elif n_n == 20:
            ideal_ratio = 1.43267
        elif n_n == 21:
            ideal_ratio = 1.49119
        elif n_n == 22:
            ideal_ratio = 1.54840
        elif n_n == 23:
            ideal_ratio = 1.60436
        else:
            ideal_ratio = 1.65915

        actual_ratio = c_r / n_e_r
        output = ideal_ratio / actual_ratio
        return output

    @classmethod
    def get_cluster_range(self, radii, packing_threshold):
        """Function compute the maximum and minimum possible cluster sizes,
        given a list of radii.

        The smallest possible cluster has the smallest
        atom in the center and the largest in the outside. The largest
        possible has the largest in the inside and the smallest in the outside.

        Parameters
        ----------
        radii : array-like
            Radii of elements in the system. A list of float values.
        packing_threshold : float
            APE defining maximum packing threshold.

        Returns
        ----------
        min_cluster_size : int
            Minimum cluster size as defined by number of atoms in the shell.
        max_cluster_size : int
            Maximum cluster size as defined by number of atoms in the shell.
        """

        l_r = len(radii)

        # Get the indices of the maximum and minimum radius.
        biggest_radius = int(np.argmax(radii))
        smallest_radius = int(np.nanargmin(radii))

        # Compute the smallest possible cluster.
        cluster = np.zeros(l_r)
        center_type = smallest_radius
        cluster[biggest_radius] = 3

        while self.compute_APE(radii=radii, center_type=center_type,
                               shell_types=cluster) < (1 - packing_threshold):
            cluster[biggest_radius] += 1
            if cluster[biggest_radius] > 24:
                raise RuntimeError("Smallest cluster > 24 atoms: "
                                   "packing_threshold must be too large")

        smallest_cluster = cluster[biggest_radius]

        # Compute the largest possible cluster.
        cluster[biggest_radius] = 0
        cluster[smallest_radius] = 24
        center_type = biggest_radius

        while self.compute_APE(radii=radii, center_type=center_type,
                               shell_types=cluster) > (1 + packing_threshold):
            cluster[smallest_radius] -= 1
            if cluster[smallest_radius] < 3:
                raise RuntimeError("Largest cluster < 3 atoms: "
                                   "packing_threshold must be too large")

        biggest_cluster = cluster[smallest_radius]

        return int(smallest_cluster), int(biggest_cluster)

    def get_closest_compositions(self, target_composition,
                                 other_compositions, n_closest, p_norm):
        """Function to get closest compositions from a given target
        composition.

        Parameters
        ----------
        target_composition : CompositionEntry
            Composition from which to measure distance.
        other_compositions : array-like
            Compositions whose distance from the target will be ranked. A
            list of CompositionEntry's.
        n_closest : int
            Number of closest compounds to return.
        p_norm : int
            P-norm to use when computing distance.

        Returns
        ----------
        dist : array-like
            Distances of the compositions from the target composition. A list
            of float values.
        comps : array-like
            A list of CompositionEntry's.

        """

        tmp_list = []
        for comp in other_compositions:
            cdf = -CompositionDistanceFilter.compute_distance(comp,
                  target_composition, p_norm)
            heappush(tmp_list, (cdf, comp))
            if len(tmp_list) > n_closest:
                heappop(tmp_list)

        comps = []
        dist = []
        while tmp_list:
            tup = heappop(tmp_list)
            dist.insert(0, -tup[0])
            comps.insert(0, tup[1])
        return dist, comps

    @classmethod
    def find_efficiently_packed_clusters(self, radii, packing_threshold):
        """Function to find all clusters with better APE than a certain
        threshold, given a list of atomic radii.

        The packing efficiency is defined as abs(1 - APE).

        Parameters
        ----------
        radii : array-like
            Radii of elements. A list of float values.
        packing_threshold : float
            Desired packing limit threshold. A "default" choice would be 0.05.

        Returns
        ----------
        output : array-like
            A list of efficiently packed structures for each atom type as the
            central atom. Ex: x[0][1] is the 2nd efficiently packed cluster
            with atom type 0 as the central atom. A list containing a list of
            int values.

        """

        output = []
        l_r = len(radii)

        # Special case: Only one atom type.
        if l_r == 1:
            clusters =[]

            # Loop through all cluster sizes.
            for i in xrange(3, 24):
                ape = self.compute_APE(i, center_radius=1.0,
                                       neigh_eff_radius=1.0)
                if abs(ape - 1) < 0.05:
                    clusters.append([i])

            output.append(clusters)
            return output

        # Determine the minimum and maximum cluster sizes.
        min_cluster_size, max_cluster_size = self.get_cluster_range(radii,
                                            packing_threshold)

        esc = EqualSumCombinations(max_cluster_size - 1, l_r)
        for cluster_size in xrange(min_cluster_size,
                                   max_cluster_size):
            if not esc.dp[cluster_size][l_r]:
                shells = esc.get_combinations(cluster_size, l_r)

        # Loop through each atom as the central type.
        for central_type in xrange(l_r):
            clusters = []

            # Loop over possible ranges of cluster sizes (determined from
            # radii).
            for cluster_size in xrange(min_cluster_size, max_cluster_size):

                # Loop through all combinations of atom types in the first
                # shell.
                for shell in esc.dp[cluster_size][l_r]:
                    ape = self.compute_APE(radii=radii,
                                           center_type=central_type,
                                           shell_types=shell)

                    if abs(ape - 1) < packing_threshold:
                        clusters.append(list(shell))

            output.append(clusters)

        return output

    @classmethod
    def compute_cluster_compositions(self, e_ids, clusters):
        """Function to compute the compositions of a list of atomic clusters.

        The composition includes both atoms in the first nearest neighbor
        shell and the atom in the center of the cluster.

        Parameters
        ----------
        e_ids: array-like
            Ids of the elements from which clusters are composed. A list of
            int values.
        clusters : array-like
            Clusters to convert. List of identity shell compositions for each
            type of central atom. Ex: clusters[1][2] is an array defining the
            number of atoms of each type for clusters with an atom of type 1
            in the center. A list containing a list of int values.

        Returns
        ----------
        output : array-like
            Compositions found in this cluster. A list of CompositionEntry's.

        """

        output = []
        l_c = len(clusters)

        # Loop through clusters with each type of atom at the center.
        for ct in xrange(l_c):
            for shell in clusters[ct]:
                fractions = list(shell)
                fractions[ct] += 1.0

                entry = CompositionEntry(element_ids=e_ids, fractions=fractions)
                output.append(entry)

        return output

    @classmethod
    def determine_optimal_APE(self, central_atom_type, shell_composition,
                              radii):
        """Function to compute the optimal APE for a cluster with a certain
        atom type in the center and composition in the cell.


        Parameters
        ----------
        central_atom_type : int
            Element id (Z - 1) of th central atom.
        shell_composition : CompositionEntry
            Composition of the nearest-neighbor shell.
        radii : array-like
            Lookup table of elemental radii. A list of float values.

        Returns
        ----------
        output : float
            The optimal APE.

        Notes
        -----
        This algorithm finds the number of atoms in the shell such that the
        APE of the cluster is closest to 1. Note: This calculation assumes
        that sites in the first nearest-neighbor shell can be
        partially-occupied.

        """

        # Initialize output.
        output = float("inf")

        # Get radius of center, mean radius of outside.
        center_r = radii[central_atom_type]
        tmp_r = [radii[elem] for elem in shell_composition.get_element_ids()]
        shell_r = np.average(tmp_r,
                             weights=shell_composition.get_element_fractions())

        # Loop through all atom sizes.
        for z in xrange(3, 24):
            ape = self.compute_APE(n_neighbors=z,
                                   center_radius=center_r,
                                   neigh_eff_radius=shell_r)
            if abs(ape - 1) < abs(output - 1):
                output = ape
            else:
                break

        return output

    def generate_features(self, entries):
        """Function to generate features as mentioned in the class description.

        Parameters
        ----------
        entries : array-like
            Compositions for which features are to be generated. A list of
            CompositionEntry's.

        Returns
        ----------
        features : DataFrame
            Features for the given entries. Pandas data frame containing the
            names and values of the descriptors.

        Raises
        ------
        ValueError
            If input is not of type list.
            If items in the list are not CompositionEntry instances.

        """

        # Initialize lists of feature values and headers for pandas data frame.
        feat_values = []
        feat_headers = []

        # Raise exception if input argument is not of type list of
        # CompositionEntry's.
        if (type(entries) is not types.ListType):
            raise ValueError("Argument should be of type list of "
                             "CompositionEntry's")
        elif (entries and not isinstance(entries[0], CompositionEntry)):
            raise ValueError("Argument should be of type list of "
                             "CompositionEntry's")

        # Get the atomic radii.
        radii_lookup = LookUpData.load_property(self.radius_property)

        # Insert header names here.
        for n in self.n_nearest_to_eval:
            feat_headers.append("APE_Nearest{}_Below{}".format(n,
                                        self.packing_threshold))

        feat_headers.append("APE_SystemAverage")
        feat_headers.append("APE_SystemAverageDeviation")

        # Find the largest number of clusters to be considered.
        largest_n = max(self.n_nearest_to_eval)

        # Get the entries, sort so that alloys for the same system are
        # grouped together.
        entries.sort()

        # Compute features for each entry.
        last_elements = []
        clusters = []
        for entry in entries:
            # print entry
            tmp_list = []
            cur_elements = list(entry.get_element_ids())
            cur_fractions = list(entry.get_element_fractions())

            # If list of elements is greater than max_n_types, pick only the
            # most prevalent.
            if len(cur_elements) > self.max_n_types:
                cur_elements = [x for _,x in sorted(zip(cur_fractions,
                                        cur_elements), reverse=True)]
                cur_elements = cur_elements[:self.max_n_types]
                # cur_elements = sorted(cur_elements, key=entry.get)[
                #                :self.max_n_types]

            # Sort elements by atomic number.
            cur_elements.sort()

            # Get radii of those elements.
            radii = [radii_lookup[e_id] for e_id in cur_elements]

            # If current element list doesn't equal last_elements, recompute
            # list of nearest clusters.
            if cur_elements != last_elements:
                tmp = self.find_efficiently_packed_clusters(radii,
                                                        self.packing_threshold)
                clusters = self.compute_cluster_compositions(cur_elements, tmp)
                last_elements = cur_elements

            # Find the closest clusters and distance to our cluster.
            distances, closest_clusters = self.get_closest_compositions(entry,
                                            clusters, largest_n, 2)
            l_d = len(distances)
            if l_d == 0:
                for n in self.n_nearest_to_eval:
                    tmp_list.append(1000.0)
            else:
                for n in self.n_nearest_to_eval:
                    tmp_list.append(np.mean(distances[:min(n, l_d - 1)]))

            # Compute the packing efficiency of clusters around each atom
            # assuming that the composition of the first nearest-neighbor
            # shell is equal to the composition of the alloy.

            entry_elements = list(entry.get_element_ids())
            entry_fractions = list(entry.get_element_fractions())
            cluster_APEs = [self.determine_optimal_APE(id, entry,
                                    radii_lookup) for id in entry_elements]

            # Compute the composition-weighted average and average deviation
            # from 1.
            avg = np.average(cluster_APEs, weights=entry_fractions)
            avg_dev = np.average([abs(1.0 - c) for c in cluster_APEs],
                                 weights=entry_fractions)
            tmp_list.append(avg)
            tmp_list.append(avg_dev)

            feat_values.append(tmp_list)

        features = pd.DataFrame(feat_values, columns=feat_headers)
        return features


import math
import types
import pandas as pd
from ....data.materials.CompositionEntry import CompositionEntry
from ....data.materials.util.LookUpData import LookUpData
from ....utility.tools.OxidationStateGuesser import OxidationStateGuesser

class IonicityAttributeGenerator:
    """Class to generate the attributes based on the ionicity of a compound.

    Creates attributes based on whether it is possible to form a
    charge-neutral ionic compound, and two features based on a simple measure
    of "bond ionicity" (see Ref. [1]).

    Notes
    -----
    Bond ionicity is defined as:
    .. math:: I(x,y) = 1 - \exp(-0.25* [\chi(x) - \chi(y)]^2)

    Maximum ionic character: Max I(x,y) between any two constituents.
    Mean ionic character: :math: `\sum x_i * x_j * I(i,j)` where :math: `x_i`
    is the fraction of element :math: `i` and :math: `\chi(x)` is the
    electronegativity of element :math: `x`.

    References
    ----------
    .. [1] William D. Callister, Materials science and engineering: an
    introduction, Hoboken: John Wiley, 2014.

    """

    def generate_features(self, entries):
        """Function to generate features as mentioned in the class description.

        Parameters
        ----------
        entries : array-like
            Compositions for which features are to be generated. A list of
            CompositionEntry's.

        Returns
        ----------
        features : DataFrame
            Features for the given entries. Pandas data frame containing the
            names and values of the descriptors.

        Raises
        ------
        ValueError
            If input is not of type list.
            If items in the list are not CompositionEntry instances.

        """

        # Initialize lists of feature values and headers for pandas data frame.
        feat_values = []
        feat_headers = []

        # Raise exception if input argument is not of type list of
        # CompositionEntry's.
        if (type(entries) is not types.ListType):
            raise ValueError("Argument should be of type list of "
                             "CompositionEntry's")
        elif (entries and not isinstance(entries[0], CompositionEntry)):
            raise ValueError("Argument should be of type list of "
                             "CompositionEntry's")

        # Insert header names here.
        feat_headers.append("CanFormIonic")
        feat_headers.append("MaxIonicChar")
        feat_headers.append("MeanIonicChar")

        # Instantiate and initialize oxidation state guesser with
        # electronegativity and oxidation state values.
        ox_guesser = OxidationStateGuesser()
        en = LookUpData.load_property("Electronegativity")
        ox_guesser.set_electronegativity(en)
        ox_guesser.set_oxidationstates(LookUpData.load_property(
            "OxidationStates"))

        for entry in entries:
            tmp_list = []
            # Can it form an ionic compound?
            tmp_list.append(0 if len(ox_guesser.get_possible_states(entry))
                                 == 0 else 1)
            tmp_en = []

            # Compute mean ionic character.
            mean_ionic = 0.0
            elems = entry.get_element_ids()
            fracs = entry.get_element_fractions()
            for i,elem1 in enumerate(elems):
                tmp_en.append(en[elem1])
                for j,elem2 in enumerate(elems):
                    m = 1 - math.exp(-0.25*(en[elem1] - en[elem2])**2)

                    mean_ionic += fracs[i] * fracs[j] * m

            # Compute max ionic character.
            max_ionic = 1 - math.exp(-0.25 * (max(tmp_en) - min(tmp_en)) ** 2)
            tmp_list.append(max_ionic)
            tmp_list.append(mean_ionic)
            feat_values.append(tmp_list)

        features = pd.DataFrame(feat_values, columns=feat_headers)
        return features


# coding=utf-8
import math
import types
import numpy as np
import pandas as pd
from ....data.materials.CompositionEntry import CompositionEntry
from ....data.materials.util.LookUpData import LookUpData

class YangOmegaAttributeGenerator:
    """Class to compute the attributes :math:`\Omega` and :math:`\delta`
    developed by Yang and Zhang [1].

    These parameters are based on the liquid formation enthalpy and atomic
    sizes of elements respectively and were originally developed to predict
    whether a metal alloy will form a solid solution of bulk metallic glass.

    Notes
    -----
    :math: `\Omega` is derived from the melting temperature, ideal mixing
    entropy, and regular solution solution interaction parameter (
    :math: `\Omega_{i,j}`) predicted by the Miedema model for binary liquids.
    Specifically, it is computed using the relationship:
    .. math:: \Omega = \displaystyle\frac{T_m \Delta S_{mix}} {|\Delta H_{mix}|}

    where :math: `T_m` is the composition-weighted average of the melting
    temperature, :math: `\Delta S_{mix}` is the ideal solution entropy,
    and :math: `\Delta H_{mix}` is the mixing enthalpy. The mixing enthalpy
    is computed using the Miedema mixing enthalpies tabulated by Takeuchi and
    Inoue [2] where:
    .. math:: \Delta H_{mix} = \displaystyle\sum \Omega_{i,j} c_i c_j
    and :math: `\Omega_{i,j} = 4 * \Delta H_{mix}`.

    :math: `\delta` is related to the polydispersity of atomic sizes, and is
    computed using the relationship:

    .. math:: \delta = [\displaystyle\sum c_i (1 - \frac{r_i}{r_{
    average})^2]^0.5

    where :math: `r_i` is the atomic size. Here, we use the atomic radii
    compiled by Miracle et al. [3] rather than those compiled by Kittel,
    as in the original work.

    References
    ----------
    .. [1] X. Yang and Y. Zhang, "Prediction of high-entropy stabilized
    solid-solution in multi-component alloys," Materials Chemistry and
    Physics, vol. 132, no. 2--3, pp. 233--238, Feb. 2012.
    .. [2] A. Takeuchi and A. Inoue, "Classification of Bulk Metallic Glasses
    by Atomic Size Difference, Heat of Mixing and Period of Constituent
    Elements and Its Application to Characterization of the Main Alloying
    Element," MATERIALS TRANSACTIONS, vol. 46, no. 12, pp. 2817--2829, 2005.
    .. [3] D. B. Miracle, D. V. Louzguine-Luzgin, L. V. Louzguina-Luzgina,
    and A. Inoue, "An assessment of binary metallic glasses: correlations
    between structure, glass forming ability and stability," International
    Materials Reviews, vol. 55, no. 4, pp. 218--256, Jul. 2010.

    """

    def generate_features(self, entries):
        """Function to generate features as mentioned in the class description.

        Parameters
        ----------
        entries : array-like
            Compositions for which features are to be generated. A list of
            CompositionEntry's.

        Returns
        ----------
        features : DataFrame
            Features for the given entries. Pandas data frame containing the
            names and values of the descriptors.

        Raises
        ------
        ValueError
            If input is not of type list.
            If items in the list are not CompositionEntry instances.

        """

        # Initialize lists of feature values and headers for pandas data frame.
        feat_values = []
        feat_headers = []

        # Raise exception if input argument is not of type list of
        # CompositionEntry's.
        if (type(entries) is not types.ListType):
            raise ValueError("Argument should be of type list of "
                             "CompositionEntry's")
        elif (entries and not isinstance(entries[0], CompositionEntry)):
            raise ValueError("Argument should be of type list of "
                             "CompositionEntry's")

        # Insert header names here.
        feat_headers.append("Yang_Omega")
        feat_headers.append("Yang_delta")

        # Load property values here.
        radii = LookUpData.load_property("MiracleRadius")
        meltingT = LookUpData.load_property("MeltingT")
        miedema = LookUpData.load_pair_property("MiedemaLiquidDeltaHf")

        for entry in entries:
            tmp_list = []
            tmp_radii = []
            tmp_meltingT = []

            elem_fracs = entry.get_element_fractions()
            elem_ids = entry.get_element_ids()
            for elem_id in elem_ids:
                tmp_radii.append(radii[elem_id])
                tmp_meltingT.append(meltingT[elem_id])

            # Compute the average melting point.
            averageTm = np.average(tmp_meltingT, weights=elem_fracs)

            # Compute the ideal entropy.
            entropy = 0.0
            for f in elem_fracs:
                entropy += f*math.log(f) if f > 0 else 0.0
            entropy *= 8.314/1000

            # Compute the enthalpy
            enthalpy = 0.0

            for i in range(len(elem_ids)):
                for j in range(i + 1, len(elem_ids)):
                    enthalpy += miedema[max(elem_ids[i], elem_ids[j])][min(
                        elem_ids[i], elem_ids[j])] * elem_fracs[i] * \
                                elem_fracs[j]
            enthalpy *= 4

            # Compute omega
            tmp_list.append(abs(averageTm * entropy / enthalpy))

            # Compute delta
            delta_squared = 0.0
            average_r = np.average(tmp_radii, weights=elem_fracs)
            for i in range(len(elem_ids)):
                delta_squared += elem_fracs[i] * (1 - tmp_radii[i] /
                                                  average_r)**2

            tmp_list.append(math.sqrt(delta_squared))

            feat_values.append(tmp_list)

        features = pd.DataFrame(feat_values, columns=feat_headers)
        return features


import pandas as pd
import numpy as np
import types
from ....data.materials.CrystalStructureEntry import CrystalStructureEntry

class LatticeSimilarityAttributeGenerator:
    """Compute similarity of structure to several simple lattices.

    Notes
    -----
    Determined by comparing the shape of each coordination polyhedron in the
    structure (as determined using a Voronoi tessellation) to those in a
    reference lattice.

    Similarity is computed by summing the difference in the number of faces
    with each number of edges between a certain Voronoi cell and that of the
    reference lattice. This difference is then normalized by the number of
    faces in the reference lattice, and averaged over all atoms to produce a
    "similarity index". In this form, structures based on the reference
    lattice have a match of 0, which becomes larger with increase
    dissimilarity.

    For now we consider the BCC, FCC (which has the same coordination
    polyhedron shape as HCP), and SC lattices.

    """

    def generate_features(self, entries):
        """Function to generate features as mentioned in the class description.

        Parameters
        ----------
        entries : array-like
            Crystal structures for which features are to be generated. A list
            of CrystalStructureEntry's.

        Returns
        ----------
        features : DataFrame
            Features for the given entries. Pandas data frame containing the
            names and values of the descriptors.

        Raises
        ------
        ValueError
            If input is not of type list.
            If items in the list are not CrystalStructureEntry instances.

        """

        # Raise exception if input argument is not of type list of
        # CrystalStructureEntry's.

        if (type(entries) is not types.ListType):
            raise ValueError("Argument should be of type list of "
                             "CrystalStructureEntry's")
        elif (entries and not isinstance(entries[0], CrystalStructureEntry)):
            raise ValueError("Argument should be of type list of "
                             "CrystalStructureEntry's")

        feat_headers = []
        feat_values = []

        feat_headers.append("dissimilarity_FCC")
        feat_headers.append("dissimilarity_BCC")
        feat_headers.append("dissimilarity_SC")

        l_fh = len(feat_headers)
        for entry in entries:
            temp_list = []
            try:
                output = entry.compute_voronoi_tessellation()
            except Exception:
                tmp_list = [np.nan] * l_fh # If tessellation fails.
                feat_values.append(tmp_list)
                continue

            fcc = output.mean_fcc_dissimilarity()
            bcc = output.mean_bcc_dissimilarity()
            sc = output.mean_sc_dissimilarity()

            temp_list.append(fcc)
            temp_list.append(bcc)
            temp_list.append(sc)

            feat_values.append(temp_list)

        features = pd.DataFrame(feat_values, columns=feat_headers)
        return features


#!/usr/bin/env python

import sys
import os
import time
import copy
import inspect

from .. import __version__, __author__
from .pandas_pd import pdw
from .cheml_cml import cmlw
from .sklearn_skl import sklw
from .tensorflow_tf import tfw

from ..utils import isint, value, std_datetime_str, tot_exec_time_str
from .base import LIBRARY

def banner():
    PROGRAM_NAME = "ChemML"
    PROGRAM_VERSION = __version__
    REVISION_DATE = "2018-03-20"
    AUTHORS = __author__
    CONTRIBUTORS = " "
    DESCRIPTION = "ChemML is a machine learning and informatics program suite for the chemical and materials sciences."
    str = []
    str.append("=================================================")
    str.append(PROGRAM_NAME + " " + PROGRAM_VERSION + " (" + REVISION_DATE + ")")
    for AUTHOR in AUTHORS:
        str.append(AUTHOR)
    str.append("=================================================")
    str.append(time.ctime())
    str.append("")
    # str.append(DESCRIPTION)
    # str.append("")

    print
    for line in str:
        print line

class Parser(object):
    """
    script: list of strings
        A list of lines in the cheml script file.
    """
    def __init__(self, script):
        self.script = script

    def fit(self):
        """
        The main funtion for parsing cheml script.
        It starts with finding blocks and then runs other functions.

        :return:
        cmls: cheml script
        """
        blocks={}
        it = -1
        check_block = False
        for line in self.script:
            if '##' in line:
                it += 1
                blocks[it] = [line]
                check_block = True
                continue
            elif '#' in line:
                check_block = False
            elif check_block and ('<' in line or '>' in line):
                blocks[it].append(line)
                continue

        cmls = self._options(blocks)
        ImpOrder,CompGraph = self.transform(cmls)
        self._print_out(cmls)
        return cmls, ImpOrder, CompGraph

    def _functions(self, line):
        if '<' in line:
            function = line[line.index('##')+2:line.index('<')].strip()
        elif '>' in line:
            function = line[line.index('##')+2:line.index('>')].strip()
        else:
            function = line[line.index('##')+2:].strip()
        return function

    def _parameters(self, block,item):
        parameters = {}
        send = {}
        recv = {}
        for line in block:
            while '<<' in line:
                line = line[line.index('<<')+2:].strip()
                if '<' in line:
                    args = line[:line.index('<')].strip()
                else:
                    args = line.strip()
                param = args[:args.index('=')].strip()
                val = args[args.index('=')+1:].strip()
                parameters[param] = value(val) #val #"%s"%val
            while '>>' in line:
                line = line[line.index('>>') + 2:].strip()
                if '>' in line:
                    args = line[:line.index('>')].strip()
                else:
                    args = line.strip()
                arg = args.split()
                if len(arg) == 2:
                    a, b = arg
                    if isint(b) and not isint(a):
                        send[(a, int(b))] = item
                    elif isint(a) and not isint(b):
                        recv[(b, int(a))] = item
                    else:
                        msg = 'wrong format of send and receive in block #%i at %s (send: >> var id; recv: >> id var)' % (item+1, args)
                        raise IOError(msg)
                else:
                    msg = 'wrong format of send and receive in block #%i at %s (send: >> var id; recv: >> id var)'%(item+1,args)
                    raise IOError(msg)
        return parameters, send, recv

    def _options(self, blocks):
        cmls = []
        for item in xrange(len(blocks)):
            block = blocks[item]
            function = self._functions(block[0])
            parameters, send, recv = self._parameters(block,item)
            cmls.append({"task": function,
                         "parameters": parameters,
                         "send": send,
                         "recv": recv})
        return cmls

    def _print_out(self, cmls):
        item = 0
        for block in cmls:
            item+=1
            line = '%s\n' %(block['task'])
            line = line.rstrip("\n")
            tmp_str =  '%i'%item+' '*(4-len(str(item)))+'Task: '+line
            print tmp_str
            line = '<<<<<<<'
            line = line.rstrip("\n")
            tmp_str = '        ' + line
            print tmp_str
            if len(block['parameters']) > 0 :
                for param in block['parameters']:
                    line = '%s = %s\n'%(param,block['parameters'][param])
                    line = line.rstrip("\n")
                    tmp_str =  '        '+line
                    print tmp_str
            else:
                line = ' :no parameter passed: set to default values if available'
                line = line.rstrip("\n")
                tmp_str =  '        ' + line
                print tmp_str
            line = '>>>>>>>'
            line = line.rstrip("\n")
            tmp_str =  '        ' + line
            print tmp_str
            if len(block['send']) > 0:
                for param in block['send']:
                    line = '%s -> send (id=%i)\n' %(param[0],param[1])
                    line = line.rstrip("\n")
                    tmp_str =  '        ' + line
                    print tmp_str
            else:
                line = ' :nothing to send:'
                line = line.rstrip("\n")
                tmp_str =  '        ' + line
                print tmp_str
            if len(block['recv']) > 0:
                for param in block['recv']:
                    line = '%s <- recv (id=%i)\n' %(param[0],param[1])
                    line = line.rstrip("\n")
                    tmp_str = '        ' + line
                    print tmp_str
            else:
                line = ' :nothing to receive:'
                line = line.rstrip("\n")
                tmp_str = '        ' + line
                print tmp_str
            line = ''
            line = line.rstrip("\n")
            tmp_str = '        ' + line
            print tmp_str

    def transform(self, cmls):
        """
        goals:
            - collect all sends and receives
            - check send and receive format.
            - make the computational graph
            - find the order of implementation of functions based on sends and receives

        :param cmls:
        :return implementation order:
        """
        send_all = []
        recv_all = []
        for block in cmls:
            send_all += block['send'].items()
            recv_all += block['recv'].items()
        # check send and recv
        if len(send_all) > len(recv_all):
            msg = '@cheml script - number of sent tokens must be less or equal to number of received tokens'
            raise ValueError(msg)
        send_ids = [k[1] for k,v in send_all]
        recv_ids = [k[1] for k,v in recv_all]
        for id in send_ids:
            if send_ids.count(id)>1:
                msg = 'identified non unique send id (id#%i)'%id
                raise NameError(msg)
        if set(send_ids) != set(recv_ids):
            print set(send_ids),set(recv_ids)
            msg = 'missing pairs of send and receive id:\n send IDs:%s\n recv IDs:%s\n'%(str(set(send_ids)),str(set(recv_ids)))
            raise ValueError(msg)

        # make graph
        reformat_send = {k[1]:[v,k[0]] for k,v in send_all}
        CompGraph = tuple([tuple(reformat_send[k[1]]+[v,k[0]]) for k,v in recv_all])

        # find orders
        ids_sent = []
        ImpOrder = []
        inf_checker = 0
        while len(ImpOrder)<len(cmls):
            inf_checker +=1
            for i in range(len(cmls)):
                if i not in ImpOrder:
                    ids_recvd = [k[1] for k,v in cmls[i]['recv'].items()]
                    if len(ids_recvd) == 0:
                        ids_sent += [k[1] for k,v in cmls[i]['send'].items()]
                        ImpOrder.append(i)
                    elif len(set(ids_recvd) - set(ids_sent)) == 0:
                        ids_sent += [k[1] for k,v in cmls[i]['send'].items()]
                        ImpOrder.append(i)
            if  inf_checker > len(cmls):
                msg = 'Your design of send and receive tokens makes a loop of interdependencies. You can avoid such loops by designing your workflow hierarchichally'
                raise IOError(msg)
        return tuple(ImpOrder),CompGraph

class BASE(object):
    def __init__(self, CompGraph):
        self.graph = CompGraph
        self.graph_info = {}
        self.send = {}      # {(iblock,token):output class}
        self.requirements = ['pandas']
        self.start_time = time.time()
        self.block_time = 0
        self.date = std_datetime_str('date')
        self.time = std_datetime_str('time')
        self.InputScript = ''
        self.output_directory = '.'
        self.log = []

class Wrapper(LIBRARY):
    """
    Todo: documentation
    """
    def __init__(self, cmls, ImpOrder, CompGraph, InputScript, output_directory):
        self.Base = BASE(CompGraph)     # initial and only instance of BASE during the entire wrapper run
        self.Base.InputScript = InputScript
        self.Base.output_directory = output_directory
        self.ImpOrder = ImpOrder
        self.cmls = cmls
        tmp_str = "=================================================\n"
        print tmp_str
        self._checker()

    def _checker(self):
        """
        check for key parameters like module and function and
        any possible typo in other params.
        """
        # get params
        # legal_tasks = ['DataRepresentation','Script','Input','Output','Preprocessor','FeatureSelection',
        #                         'FeatureTransformation','Divider','Regression','Classification','Postprocessor',
        #                         'Evaluation','Visualization','Optimizer']
        # legal_tasks = ['Enter','Prepare','Model','Search','Mix','Visualize','Store']
        # run over graph
        for iblock, block in enumerate(self.cmls):
            # check super function
            task = block['task']
            # if task not in legal_tasks:
            #     msg = '@Task #%i(%s): %s is not a valid task' %(iblock + 1, task,task)
            #     raise NameError(msg)
            # check parameters
            parameters = block['parameters']
            if 'host' not in parameters:
                msg = "@Task #%i(%s): no host name found" % (iblock + 1, task)
                raise NameError(msg)
            if 'function' not in parameters:
                msg = "@Task #%i(%s): no function name found" % (iblock + 1, task)
                raise NameError(msg)
            # check host and function
            # host_function = (block['parameters']['host'], block['parameters']['function'])
        return 'The input file is in a correct format.'

    def call(self):
        self.refs = {}
        for iblock in self.ImpOrder:
            task = self.cmls[iblock]['task']
            parameters = self.cmls[iblock]['parameters']
            host = parameters.pop('host')
            function = parameters.pop('function')
            start_time = time.time()
            tmp_str =  "======= block#%i: (%s, %s)" % (iblock + 1, host, function)
            print tmp_str
            tmp_str = "| run ...\n"
            print tmp_str
            if host == 'sklearn':
                # check methods
                legal_functions = [klass[0] for klass in inspect.getmembers(sklw)]
                if function in legal_functions:
                    self.references(host, function)  # check references
                    self.Base.graph_info[iblock] = (host, function)
                    cml_interface = [klass[1] for klass in inspect.getmembers(sklw) if klass[0] == function][0]
                    cmli = cml_interface(self.Base, parameters, iblock, task, function, host)
                    cmli.run()
                else:
                    self.references(host, function)  # check references
                    self.Base.graph_info[iblock] = (host, function)
                    cml_interface = [klass[1] for klass in inspect.getmembers(sklw) if klass[0] == 'automatic_run'][0]
                    cmli = cml_interface(self.Base, parameters, iblock, task, function, host)
                    cmli.run()
            elif host == 'cheml':
                # check methods
                legal_functions = [klass[0] for klass in inspect.getmembers(cmlw)]
                if function not in legal_functions:
                    msg = "@function #%i: couldn't find function '%s' in the module '%s' wrarpper" %(iblock,function,host)
                    raise NameError(msg)
                self.references(host,function) # check references
                self.Base.graph_info[iblock] = (host, function)
                cml_interface = [klass[1] for klass in inspect.getmembers(cmlw) if klass[0] == function][0]
                cmli = cml_interface(self.Base, parameters, iblock,task,function,host)
                cmli.run()
            elif host == 'pandas':
                # check methods
                legal_functions = [klass[0] for klass in inspect.getmembers(pdw)]
                if function not in legal_functions:
                    msg = "@function #%i: couldn't find function '%s' in the module '%s' wrarpper" %(iblock,function,host)
                    raise NameError(msg)
                self.references(host,function) # check references
                self.Base.graph_info[iblock] = (host, function)
                cml_interface = [klass[1] for klass in inspect.getmembers(pdw) if klass[0] == function][0]
                cmli = cml_interface(self.Base, parameters, iblock,task,function,host)
                cmli.run()
            elif host == 'tensorflow':
                # check methods
                legal_functions = [klass[0] for klass in inspect.getmembers(tfw)]
                if function not in legal_functions:
                    msg = "@function #%i: couldn't find function '%s' in the module '%s' wrarpper" %(iblock,function,host)
                    raise NameError(msg)
                self.references(host,function) # check references
                self.Base.graph_info[iblock] = (host, function)
                cml_interface = [klass[1] for klass in inspect.getmembers(tfw) if klass[0] == function][0]
                cmli = cml_interface(self.Base, parameters, iblock,task,function,host)
                cmli.run()

            end_time = tot_exec_time_str(start_time)
            tmp_str = "| ... done!"
            print tmp_str
            tmp_str = '| '+end_time
            print tmp_str
            tmp_str = "=======\n\n"
            print tmp_str
        self._save_references()
        tmp_str = "Total " + tot_exec_time_str(self.Base.start_time)
        print tmp_str
        tmp_str = std_datetime_str() + '\n'
        print tmp_str

class Settings(object):
    """
    makes the output directory.

    Parameters
    ----------
    output_directory: String, (default = "CheML.out")
        The directory path/name to store all the results and outputs

    input_copy: Boolean, (default = True)
        If True, keeps a copy of input script in the output_directory

    Returns
    -------
    output_directory
    """
    def __init__(self,output_directory="CMLWrapper.out"):
        self.output_directory = output_directory

    def fit(self):
        initial_output_dir = copy.deepcopy(self.output_directory)
        i = 0
        while os.path.exists(self.output_directory):
            i+=1
            self.output_directory = initial_output_dir + '%i'%i
        os.makedirs(self.output_directory)
        logfile = open(self.output_directory + '/log.txt', 'a', 0)
        errorfile = open(self.output_directory + '/error.txt', 'a', 0)
        return self.output_directory, logfile, errorfile

    def write_InputScript(self,InputScript):
        with open(self.output_directory + '/InputScript.txt','w') as f:
            for line in InputScript:
                f.write("%s\n"%line)

class Logger(object):
    def __init__(self,logfile):
        self.terminal = sys.stdout
        self.log = logfile

    def write(self, message):
        self.terminal.write(message)
        self.log.write(message)

    def flush(self):
        pass

class Error(object):
    def __init__(self,errorfile):
        self.terminal = sys.stderr
        self.err = errorfile

    def write(self, message):
        self.terminal.write(message)
        self.err.write(message)

    def flush(self):
        pass


def run(INPUT_FILE, OUTPUT_DIRECTORY):
    """
    this is the only callable method in this file
    :param INPUT_FILE: path to the input file or the actual input script
    :return:
    """
    try:
        script = open(INPUT_FILE, 'r')
        script = script.readlines()
        tmp_str = "parsing the input file: %s ..."%INPUT_FILE
    except:
        if isinstance(INPUT_FILE, list):
            script = INPUT_FILE
            tmp_str = "parsing the input file: received as a list of lines ..."
        elif isinstance(INPUT_FILE, str):
            if '##' in INPUT_FILE and '>>' in INPUT_FILE:
                script = INPUT_FILE.split('\n')
                tmp_str = "parsing the input file: received in string format ..."
            else:
                banner()
                msg = "couldn't find the input file path or the input script is not valid"
                raise IOError(msg)
    settings = Settings(OUTPUT_DIRECTORY)
    OUTPUT_DIRECTORY, logfile, errorfile= settings.fit()
    sys.stdout = Logger(logfile)
    sys.stderr = Error(errorfile)
    banner()
    print tmp_str + '\n'
    settings.write_InputScript(script)
    cmls, ImpOrder, CompGraph = Parser(script).fit()
    # print cmls
    # print ImpOrder
    # print CompGraph
    # sys.exit('this is how much you get till now!')
    wrapper = Wrapper(cmls, ImpOrder, CompGraph, INPUT_FILE, OUTPUT_DIRECTORY)
    wrapper.call()


#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*
"""*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*
 
 									  ChemML PySCRIPT

#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#"""
#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*

if __name__=="__main__":
    sys.exit()



"""
The cheml.nn.keras module includes (please click on links adjacent to function names for more information):
    - MLP: :func:`~cheml.nn.keras.MLP`
    - MLP_sklearn: :func:`~cheml.nn.keras.MLP_sklearn`
"""


from .mlp import MLP
from .mlp import MLP_sklearn




__all__ = [
    'MLP',
    'MLP_sklearn'
    ]


from collections import OrderedDict
import numpy as np
from ..data.Atom import Atom
from ..data.Cell import Cell

class VASP5IO:
    """Class to read/write from VASP 5 format. Note that this version expects
    that the sixth line to contain the names of elements.

    """

    @classmethod
    def parse_file(self, file_name=None, list_of_lines=None):
        """Function to read a structure from a file or list of lines.

        Parameters
        ----------
        file_name : str
            Path to file. (Default value = None)
        list_of_lines : array-like
            Lines describing structure. (Default value = None)

        Returns
        -------
        output : Cell
            Structure as Cell.

        Raises
        ------
        Exception
            If something is wrong with input.
            If there is an error parsing the basis.
            If there is an error parsing atom types.
            If there is an error determining whether atoms are in cartesian
            units.
        """

        # Open file.
        lines = None
        if list_of_lines is None:
            with open(file_name, 'r') as f:
                # Add contents to file.
                lines = [line.strip() for line in f.readlines()]
        else:
            lines = [l.strip() for l in list_of_lines]

        if lines is None:
            raise Exception("Something went wrong. Check input.")

        # Get basis.
        factor = float(lines[1])
        basis = np.zeros((3, 3))
        try:
            for i in range(3):
                words = lines[2 + i].split()
                for j in range(3):
                    # VASP's basis is transposed so that each row is a
                    # lattice vector. See:
                    # http://cms.mpi.univie.ac.at/vasp/guide/node59.html
                    basis[j][i] = float(words[j]) * factor
        except Exception:
            raise Exception("Error parsing basis.")

        # Read atom types.
        types = None
        type_count = None
        try:
            types = lines[5].split()
            words = lines[6].split()
            type_count = [int(words[i]) for i in range(len(types))]
        except Exception:
            raise Exception("Error parsing atom types.")

        # Read middle section.
        atom_start = None
        cartesian = None
        try:
            # Check whether SD is on the see where atom positions start.
            atom_start = 7 if lines[7].lower().startswith("sele") else 8

            # See whether coordinates are in direct or cartesian units.
            cartesian = lines[atom_start - 1].lower().startswith("c")
        except Exception:
            raise Exception("Error determining whether atoms are in cartesian "
                            "units.")

        # Make the cell.
        structure = Cell()
        structure.set_basis(basis=basis)

        # Get atom positions.
        try:
            for t in range(len(types)):
                for ti in range(type_count[t]):
                    # Read position.
                    x = [float(w) for w in lines[atom_start].split()]
                    atom_start += 1
                    if cartesian:
                        x = structure.convert_cartesian_to_fractional(x)
                    atom = Atom(x, t)
                    structure.add_atom(atom)
                structure.set_type_name(t, types[t])
        except Exception:
            raise Exception("Error parsing atoms.")

        return structure

    def convert_structure_to_string(self, structure):
        """Function to convert structure to string representation - list of
        lines.

        Parameters
        ----------
        structure : Cell
            Desired structure to be converted.

        Returns
        -------
        output : array-like
            String representation - list of lines.

        """
        output = []

        # Write header.
        output.append("Automatically-generated POSCAR")
        output.append("1.0")

        # Write lattice vectors.
        lat_vectors = structure.get_lattice_vectors()
        for i in range(3):
            output.append(" {0:.10f} {1:.10f} {2:.10f}".format(lat_vectors[
                i][0], lat_vectors[i][1], lat_vectors[i][2]))

        # Gather atoms.
        names = OrderedDict()
        atoms = OrderedDict()
        for atom in structure.get_atoms():
            type = atom.get_type()
            if type in names:
                atoms[type].append(atom)
            else:
                names[type] = structure.get_type_name(type)
                atoms[type] = [atom]

        # Print atom types.
        name_line = ""
        count_line = ""
        for key in names:
            name_line += " "+str(key)
            count_line += " "+str(len(atoms[key]))

        output.append(name_line)
        output.append(count_line)

        output.append("Direct")

        # Print atoms.
        for key in atoms:
            for atom in atoms[key]:
                x = atom.get_position()
                output.append("{0:.10f} {1:.10f} {2:.10f}".format(x[0], x[1],
                                                                  x[2]))
        return output

    def write_structure_to_file(self, structure, filename):
        """Function to write a structure to a file.

        Parameters
        ----------
        structure : Cell
            Desired structure as Cell.
        filename : str
            Desired file name.

        """

        # Generate description.
        to_write = self.convert_structure_to_string(structure)

        # Write to file.
        with open(filename, 'w') as f:
            f.writelines(to_write)

    def print_structure(self, structure):
        """Function to print the structure to standard out.

        Parameters
        ----------
        structure : Cell
            Desired structure as Cell.

        Returns
        -------
        output : array-like
            String representation - list of lines.

        """

        # Get the structure as a list of strings.
        lines = self.convert_structure_to_string(structure)
        output = "\n".join(lines)
        return output


from ...data.materials.util.LookUpData import LookUpData
from ...data.utilities.generators.PhaseDiagramCompositionEntryGenerator \
    import PhaseDiagramCompositionEntryGenerator
from ...utility.tools.OxidationStateGuesser import OxidationStateGuesser

class IonicCompoundFinder:
    """Class to find nearby compositions from a given nominal composition that
    can be charge neutral.

    Works by finding all combinations of elements in
    the supplied composition that are within a certain distance of the target
    composition with less than a certain number of atoms per unit cell.
    Distance is computed as the L_1 distance of the composition vector.
    Example: Fe3Al and FeAl are 0.5 apart.

    Attributes
    ----------
    nominal_composition : CompositionEntry
        Nominal composition.
    maxmium_distance : float
        Maximum acceptable distance from nominal composition.
    max_formula_unit : int
        Maximum number of atoms in formula unit.

    """

    # Nominal composition.
    nominal_composition = None

    # Maximum acceptable distance from nominal composition.
    maximum_distance = 0.1

    # Maximum number of atoms in formula unit.
    max_formula_unit_size = 5

    def set_nominal_composition(self, entry):
        """Function to set the target composition of the ionic compound.

        Parameters
        ----------
        entry : CompositionEntry
            Desired nominal composition with element names and fractions as
            keys and values respectively.

        Returns
        -------

        """
        if len(entry.get_element_ids()) < 2:
            raise ValueError("Must be at least a binary compound.")
        self.nominal_composition = entry

    def set_maximum_distance(self, dist):
        """Function to set the allowed maximum distance from the target value.

        Note, the distance is computed as the L_1 norm of the composition
        vector assuming one of the elements is a balance (i.e., only sum the
        difference for N-1 elements).

        Parameters
        ----------
        dist : float
            Maximum allowed distance.

        """
        self.maximum_distance = dist

    def set_max_formula_unit_size(self, size):
        """Function to set maximum number of atoms in formula unit. Example:
        NaCl has 2.

        Parameters
        ----------
        size : int
            Maximum allowed size.

        """
        self.max_formula_unit_size = size

    def find_all_compounds(self):
        """Function to find all the compounds in the vicinity of the target
        composition.

        Returns
        -------
        accepted : array-like
            A list of CompositionEntry's.
        """

        # Get elements in the nominal compound.
        elems = self.nominal_composition.get_element_ids()
        fracs = self.nominal_composition.get_element_fractions()

        # Get list of all possible compositions.
        gen = PhaseDiagramCompositionEntryGenerator()
        gen.set_elements_by_index(elems)
        gen.set_even_spacing(False)
        gen.set_order(1, len(elems))
        gen.set_size(self.max_formula_unit_size)
        all_possibilities = gen.generate_entries()

        hits = []
        # Find which ones fit the desired tolerance.
        for entry in all_possibilities:
            # See if it is close enough in composition.
            dist = 0.0
            for e in range(len(elems)):
                dist += abs(fracs[e] - entry.get_element_fraction(id=elems[e]))

            if dist > self.maximum_distance:
                continue

            # See if it is ionically neutral.
            ox_g = OxidationStateGuesser()
            en = LookUpData.load_property("Electronegativity")
            os = LookUpData.load_property("OxidationStates")
            ox_g.set_electronegativity(en)
            ox_g.set_oxidationstates(os)
            can_form_ionic = len(ox_g.get_possible_states(entry)) > 0

            if can_form_ionic:
                hits.append((dist, entry))

        # Sort such that closest is first.
        hits.sort()

        # Get only compositions.
        accepted = [i[1] for i in hits]
        return accepted


"""
The 'cheml.nn' module includes neural_net, 
last modified date: April 24, 2016
"""

# from nn_psgd import mlp_hogwild
# from .nn_SGD_Shared import nn_psgd
# from .nn_SGD_Distributed import nn_dsgd_train, nn_dsgd_output
#Zfrom .nn_SGD_tensorflow import nn_tf
#from .nn_SGD_theano import nn_theano



__all__ = [

]
